import math
from copy import deepcopy

import numpy as np
from scipy.signal import find_peaks, convolve
from sigfig import round

from baseClasses import Peak, Background, StandardPeak, BoronPeak
from util import binary_search_find_nearest

#Note: Docstrings for methods used here are in baseClasses.py

class LinearBackground(Background):
    def __init__(self, slope, intercept, variances = [None, None]):
        super().__init__()
        self.slope = slope
        self.intercept = intercept
        self.variances = variances
        self.originalParams = [slope, intercept]
        self.originalVariances = variances
    @staticmethod
    def guess_params(xdata, ydata):
        lowerIndex = int(np.where(ydata == min(ydata[:len(xdata)//2]))[0][0])
        upperIndex = int(np.where(ydata == min(ydata[len(xdata)//2:]))[-1][-1])
        x1 = xdata[lowerIndex]
        y1 = ydata[lowerIndex]
        x2 = xdata[upperIndex]
        y2 = ydata[upperIndex]
        slope = float((y2-y1) / (x2 - x1))
        intercept = float(y1 - slope * x1)
        return LinearBackground(slope, intercept)
    @staticmethod
    def get_entry_fields():
        return ["Point 1 Energy", "Point 2 Energy"]

    def get_type(self):
        return "linear"

    def get_num_params(self):
        return 2
    
    #Getters and Setters
    def get_params(self):
        return [self.slope, self.intercept]
    def set_params(self, newParams):
        self.slope, self.intercept = newParams
    def get_variances(self):
        return self.variances
    def set_variances(self, variances):
        self.variances = variances
    def get_original_params(self):
        return self.originalParams
    def set_original_params(self, params):
        self.originalParams = list(params)
    def get_original_variances(self):
        return self.originalVariances
    def set_original_variances(self, variances):
        self.originalVariances = list(variances)
    
    #Model Methods
    def get_ydata(self, xdata):
        xdata = np.array(xdata)
        return self.slope * xdata + self.intercept
    
    def get_ydata_with_params(self,xdata,params):
        xdata = np.array(xdata)
        return params[0] * xdata + params[1]
    
    #I/O Methods
    def __str__(self):
        return "Linear: Slope = "+round(float(self.slope), sigfigs=4, notation='scientific')+", Intercept = "+round(float(self.intercept), sigfigs=4, notation='scientific')
    
    def handle_entry(self, entry):
        x1 = float(entry[0])
        y1 = float(entry[1])
        x2 = float(entry[2])
        y2 = float(entry[3])
        self.slope = float((y2-y1) / (x2 - x1))
        self.intercept = float(y1 - self.slope * x1)
    
class QuadraticBackground(Background):
    def __init__(self, a, b, c, variances = [None, None, None]):
        super().__init__()
        self.a = a
        self.b = b
        self.c = c
        self.variances = variances
        self.originalParams = [a, b, c]
        self.originalVariances = variances
    @staticmethod
    def guess_params(xdata, ydata):
        lowerIndex = int(np.where(ydata == min(ydata[:len(xdata)//2]))[0][0])
        upperIndex = int(np.where(ydata == min(ydata[len(xdata)//2:]))[-1][-1])
        x1 = xdata[lowerIndex]
        y1 = ydata[lowerIndex]
        x2 = xdata[upperIndex]
        y2 = ydata[upperIndex]
        b = float((y2-y1) / (x2 - x1))
        c = float(y1 - b * x1)
        return QuadraticBackground(0, b, c)

    @staticmethod
    def get_entry_fields():
        return ["Point 1 Energy", "Point 2 Energy"]

    def get_type(self):
        return "quadratic"

    def get_num_params(self):
        return 3
    
    #Getters and Setters
    def get_params(self):
        return [self.a, self.b, self.c]
    def set_params(self, newParams):
        self.a, self.b, self.c = newParams
    def get_variances(self):
        return self.variances
    def set_variances(self, variances):
        self.variances = variances
    def get_original_params(self):
        return self.originalParams
    def set_original_params(self, params):
        self.originalParams = list(params)
    def get_original_variances(self):
        return self.originalVariances
    def set_original_variances(self, variances):
        self.originalVariances = list(variances)
    
    #Model Methods
    def get_ydata(self, xdata):
        xdata = np.array(xdata)
        return self.a * xdata ** 2 + self.b * xdata + self.c
    
    def get_ydata_with_params(self,xdata,params):
        xdata = np.array(xdata)
        return params[0] * xdata ** 2 + params[1] * xdata + params[2]
    
    #I/O Methods
    def __str__(self):
        return "Quadratic: "+round(float(self.a), sigfigs=4, notation='scientific')+"x**2 + "+round(float(self.b), sigfigs=4, notation='scientific') + "x + "+round(float(self.c), sigfigs=4, notation='scientific')
    
    def handle_entry(self, entry):
        x1 = float(entry[0])
        y1 = float(entry[1])
        x2 = float(entry[2])
        y2 = float(entry[3])
        self.a = 0
        self.b = float((y2-y1) / (x2 - x1))
        self.c = float(y1 - self.a * x1)

class ArctanBackground(Background):
    """y = a * arctan(b*(x-c)) + d"""
    def __init__(self, a, b, c, variances = [None, None, None]):
        super().__init__()
        self.a = a
        self.b = b
        self.c = c
        self.variances = variances
        self.originalParams = [a, b, c]
        self.originalVariances = variances
    @staticmethod
    def guess_params(xdata, ydata):
        b = (xdata[0] + xdata[-1])/2
        lowerIndex = int(np.where(ydata == min(ydata[:len(xdata)//2]))[0][0])
        upperIndex = int(np.where(ydata == min(ydata[len(xdata)//2:]))[-1][-1])
        c = (ydata[lowerIndex] + ydata[upperIndex]) / 2
        a = -1 * (ydata[lowerIndex] - ydata[upperIndex]) / math.pi
        return ArctanBackground(a,b,c)
    @staticmethod
    def get_entry_fields():
        return ["Point 1 Energy", "Point 2 Energy"]

    def get_type(self):
        return "arctan"

    def get_num_params(self):
        return 3
    
    #Getters and Setters
    def get_params(self):
        return [self.a, self.b, self.c]
    def set_params(self, newParams):
        self.a, self.b, self.c = newParams
    def get_variances(self):
        return self.variances
    def set_variances(self, variances):
        self.variances = variances
    def get_original_params(self):
        return self.originalParams
    def set_original_params(self, params):
        self.originalParams = list(params)
    def get_original_variances(self):
        return self.originalVariances
    def set_original_variances(self, variances):
        self.originalVariances = list(variances)
    
    #Model Methods
    def get_ydata(self, xdata):
        xdata = np.array(xdata)
        return np.array([self.a * math.atan(x - self.b) + self.c for x in xdata])
    
    def get_ydata_with_params(self,xdata,params):
        xdata = np.array(xdata)
        res = np.array([params[0] * math.atan(x - params[1]) + params[2] for x in xdata])
        return res
    
    #I/O Methods
    def __str__(self):
        return "Arctan: "+round(float(self.a), sigfigs=4, notation='scientific')+"arctan(x-" +round(float(self.b), sigfigs=4, notation='scientific')+")) + " +round(float(self.c), sigfigs=4, notation='scientific')
    
    def handle_entry(self, entry):
        x1 = float(entry[0])
        y1 = float(entry[1])
        x2 = float(entry[2])
        y2 = float(entry[3])
        self.a = -1 * abs(y1 - y2) / math.pi
        self.b = (x1 + x2) / 2
        self.c = (y1 + y2) / 2
    

class GaussianPeak(StandardPeak):
    def __init__(self, ctr=0, amp=0, wid = 1, variances = []):
        super().__init__()
        self.ctr = ctr
        self.amp = amp
        self.wid = wid
        self.ctrVar = None
        self.ampVar = None
        self.widVar = None
        self.originalParams = [ctr, amp, wid]
        self.originalVariances = variances
        if variances != []:
            self.set_variances(variances)
        
    #Static Methods and Basic Info Functions
    
    @staticmethod
    def guess_params(xdata, ydata):
        step = xdata[1] - xdata[0]
        prominence = max(.15 * (max(ydata) - min(ydata)), .01)
        peaks, _ = find_peaks(ydata, width=math.floor(2/step), prominence = prominence)
        if len(peaks) == 0:
            peaks, _ = find_peaks(ydata, width=math.floor(2/step))
        return [GaussianPeak(xdata[p],ydata[p],1) for p in peaks]
    @staticmethod
    def get_entry_fields():
        return ["Center (keV): ", "Amplitude (cps): "]

    def get_type(self):
        return "gaussian"

    def get_num_params(self):
        return 3
    
    #Getters and Setters
    def set_params(self, newParams):
        self.ctr, self.amp, self.wid = newParams
    def get_params(self):
        return [self.ctr, self.amp, self.wid]
    def set_variances(self, variances):
        self.ctrVar, self.ampVar, self.widVar = variances
    def get_variances(self):
        return [self.ctrVar, self.ampVar, self.widVar]
    def get_original_params(self):
        return self.originalParams
    def set_original_params(self, params):
        self.originalParams = list(params)
    def get_original_variances(self):
        return self.originalVariances
    def set_original_variances(self, variances):
        self.originalVariances = list(variances)
    def get_ctr(self):
        return self.ctr
    
    #Model Methods

    def get_area(self):
        return self.amp * self.wid * math.sqrt(2*math.pi)

    def get_area_stdev(self):
        return self.get_area() * math.sqrt((self.ampVar/self.amp)**2+(self.widVar/self.wid)**2)
    
    def get_fwhm(self):
        return round(float(self.wid * 2 * math.sqrt(math.log(2))), decimals=2)

    def get_ydata(self, xdata):
        xdata = np.array(xdata)
        return self.amp * np.exp( -((xdata - self.ctr)/self.wid)**2)

    def get_ydata_with_params(self,xdata,params):
        xdata = np.array(xdata)
        return params[1] * np.exp( -((xdata - params[0])/params[2])**2)
    
    def handle_entry(self, entry, bounds=[0,16000]):
        self.ctr = float(entry[0])
        if self.ctr < bounds[0] or self.ctr > bounds[1]:
            raise ValueError("Out of Bounds Peak")
        self.amp = float(entry[1])
        self.wid = 1

    def __str__(self):
        return "Gaussian: Center " + str(round(float(self.ctr), decimals=1)) + " keV"

class KuboSakaiBoronPeak(BoronPeak):
    def __init__(self, E0=477.6, N0=1, D=1, delta=1, variances = []):
        super().__init__()
        self.E0 = E0
        self.N0 = N0
        self.D = D
        self.delta = delta
        self.originalParams = [E0,N0,D,delta]
        self.variances = variances
        self.originalVariances = variances

    #Static methods and basic info methods

    @staticmethod
    def guess_params(xdata, ydata):
        startIndex = binary_search_find_nearest(xdata, 477.6)
        curIndex = startIndex
        N0 = 0
        width = xdata[1] - xdata[0]
        while xdata[curIndex] < 487.6:
            N0 += 2 * width * min(ydata[curIndex], ydata[2*startIndex - curIndex])
            curIndex += 1
        return KuboSakaiBoronPeak(477.6, N0, 2, 1.3)
    
    @staticmethod
    def remove_from_data(xdata, ydata):
        newYData = deepcopy(ydata)
        startIndex = binary_search_find_nearest(xdata, 477.6)
        curIndex = startIndex
        minVal = min(ydata)
        while xdata[curIndex] < 487.6:
            toSub = min(ydata[curIndex], ydata[2*startIndex - curIndex]) - minVal
            newYData[curIndex] -= toSub
            newYData[2*startIndex - curIndex] -= toSub
            curIndex += 1
        return newYData
    
    @staticmethod
    def get_entry_fields():
        return ["Center (keV)", "Max Amplitude"]

    def get_type(self):
        return "kubo_sakai"

    def get_num_params(self):
        return 4
    
    #Getters and Setters
    def get_params(self):
        return [self.E0, self.N0, self.D, self.delta]
    def set_params(self, params):
        self.E0, self.N0, self.D, self.delta = params
    def get_variances(self):
        return self.variances
    def set_variances(self, variances):
        self.variances = variances
    def get_original_params(self):
        return self.originalParams
    def set_original_params(self, params):
        self.originalParams = list(params)
    def get_original_variances(self):
        return self.originalVariances
    def set_original_variances(self, variances):
        self.originalVariances = list(variances)
    def get_ctr(self):
        return self.E0
    
    #Model Methods
    def get_area(self):
        return self.N0

    def get_area_stdev(self):
        return math.sqrt(self.variances[1])
    
    def get_fwhm(self):
        return "N/A"

    #Equations for the next 2 functions from: doi.org/10.1007/s10967-007-0229-7, modified to use simple instead of complex gaussian IR function
    def get_ydata(self, xdata):
        return self.get_ydata_with_params(xdata, self.get_params())

    def get_ydata_with_params(self, xdata, params):
        #Constants
        #decay rate of Li-7*
        decayConstant = 9.49516685699
        #Speed of Light
        c = 3*10**8
        #Initial Velocity of Li-7*
        v0=4.8*10**6
        E0, N0, D, delta = params
        step = (xdata[1] - xdata[0]) / 10
        DoGx = np.arange(xdata[0], xdata[-1]+.01, step)
        DoG = c*N0/(2*E0*v0) * decayConstant / (decayConstant - D) * (1-(c*abs(DoGx-E0)/(E0*v0))**((decayConstant - D)/D))
        for i in range(len(DoGx)):
            if abs(DoGx[i] - E0) > 7.64:
                DoG[i] = 0
                
        numSteps = int((xdata[0] + 5)//step)
        start = xdata[0] - (numSteps * step)
        DoG = np.concatenate((np.zeros(numSteps), DoG))

        IRx = np.arange(start, 5, step)
        IR = 1/(delta * math.sqrt(2*math.pi)) * math.e ** (-.5*(IRx/delta)**2) * step

        convRes = convolve(DoG, IR)
        return convRes[len(IRx)//2 + numSteps:len(IRx)//2 + numSteps + len(DoGx):10]

    #I/O Methods

    def handle_entry(self, entry, bounds=[0,16000]):
        self.E0 = float(entry[0])
        if self.E0 < bounds[0] or self.E0 > bounds[1]:
            raise ValueError("Out of Bounds Peak")
        self.N0 = 10 * float(entry[1])
        self.D = 2
        self.delta = 1.8
    
    def __str__(self):
        return "Boron Peak, Center "+str(round(float(self.E0), decimals=1))+" keV, Area "+str(round(float(self.N0), decimals=1))

class ApproxBoronPeak(BoronPeak):
    def __init__(self, ctr=477.6, amp = 1, wid=7.5, slope=1, variances = [None, None, None, None]):
        super().__init__()
        self.ctr = ctr
        self.amp = amp
        self.wid = wid
        self.slope = slope
        self.originalParams = [ctr, amp, wid, slope]
        self.variances = variances
        self.originalVariances = variances
    @staticmethod
    def guess_params(xdata, ydata):
        pass
    @staticmethod
    def remove_from_data(xdata, ydata):
        newYData = deepcopy(ydata)
        startIndex = binary_search_find_nearest(xdata, 477.6)
        curIndex = startIndex
        minVal = min(ydata)
        while xdata[curIndex] < 492.6:
            toSub = min(ydata[curIndex], ydata[2*startIndex - curIndex]) - minVal
            newYData[curIndex] -= toSub
            newYData[2*startIndex - curIndex] -= toSub
            curIndex += 1
        return newYData
    @staticmethod
    def get_entry_fields():
        return ["Center (keV)", "Max Amplitude"]

    def get_type(self):
        return "approx_boron"

    def get_num_params(self):
        return 4
    
    #Getters and Setters
    def get_params(self):
        return [self.ctr, self.amp, self.wid, self.slope]
    def set_params(self, params):
        self.ctr, self.amp, self.wid, self.slope = params
    def get_variances(self):
        return self.variances
    def set_variances(self, variances):
        self.variances = variances
    def get_original_params(self):
        return self.originalParams
    def set_original_params(self, params):
        self.originalParams = list(params)
    def get_original_variances(self):
        return self.originalVariances
    def set_original_variances(self, variances):
        self.originalVariances = list(variances)
    def get_ctr(self):
        return self.ctr


import math
from sigfig import round
from baseClasses import Evaluator

class HBondAnalysis(Evaluator):
    """Experimental Evaluator for looking at the location of the H Peak relative to some other common ones."""
    def __init__(self, ROIs):
        self.HPeak = None
        self.AlPeak = None
        self.FPeak = None
        for r in ROIs:
            if r.get_range()[0] < 1633 and r.get_range()[1] > 1633:
                pairs = r.get_peak_pairs()
                for p in pairs:
                    if p[1].get_ele() == "F-20":
                        self.FPeak = p[0]
                        break
            
            if r.get_range()[0] < 1778 and r.get_range()[1] > 1778:
                pairs = r.get_peak_pairs()
                for p in pairs:
                    if p[1].get_ele() == "Al-28":
                        self.AlPeak = p[0]
                        break
            
            if r.get_range()[0] < 2224 and r.get_range()[1] > 2224:
                pairs = r.get_peak_pairs()
                for p in pairs:
                    if p[1].get_ele() == "H-2":
                        self.HPeak = p[0]
                        break
    @staticmethod
    def get_headings(_):
        return ["Value", "95% CI +/-", "Comparator Used", "Width Ratio (H/Al)"]
    @staticmethod
    def get_name():
        return "Hydrogen Peak Location Results"
    
    def get_results(self):
        HCtr = self.HPeak.get_ctr()
        HWidth = self.HPeak.get_params()[2]
        HVar = self.HPeak.get_variances()[0]

        AlCtr = self.AlPeak.get_ctr()
        AlWidth = self.AlPeak.get_params()[2]
        AlVar = self.AlPeak.get_variances()[0]
        return [((HCtr - AlCtr) - 444.33)*1000, (2*math.sqrt(HVar+AlVar))*1000, "Al-28", HWidth/AlWidth]

class MassSensEval(Evaluator): 
    """Standard Mass/Sensitivity Evaluator"""              
    def __init__(self, ROIs):
        self.ROIs = ROIs
    @staticmethod
    def get_name():
        return "Mass/Sensitivity Results"
    @staticmethod
    def get_headings(ROI):
        output = ROI.get_peak_pairs()[0][1].get_output()
        if output == "Peak Area (cps)":
            return ["Isotope", "Peak Centroid (keV)", "FWHM Width (keV)"] + [output, output + " St. Dev"]
        return ["Isotope", "Peak Centroid (keV)", "Peak Area (cps)", "FWHM Width (keV)"] + [output, output + " St. Dev"]

    def get_results(self):
        results = []
        for r in self.ROIs:
            for p in r.get_peak_pairs():
                peak_results = p[1].get_results(p[0].get_area(), p[0].get_area_stdev())
                output = p[1].get_output()
                if output == "Peak Area (cps)":
                    formatted_results = [round(float(peak_results[0]), uncertainty = 2 * float(peak_results[1]), sep=list)[0], round(float(peak_results[1]), sigfigs = 3)]
                    results.append([p[1].get_ele(), round(float(p[0].get_ctr()), decimals=2), p[0].get_fwhm(), *formatted_results])
                else:
                    formatted_results = [round(float(peak_results[0]), uncertainty = 2 * float(peak_results[1]), sep=list)[0], round(float(peak_results[1]), sigfigs = 3)]
                    results.append([p[1].get_ele(), round(float(p[0].get_ctr()), decimals=2), round(float(p[0].get_area()), decimals=2), p[0].get_fwhm(), *formatted_results])
        return results


from models import LinearBackground, QuadraticBackground, ArctanBackground, GaussianPeak, KuboSakaiBoronPeak

#Default user preferences, to be changed from settings tab
default_prefs = {
    "peak_type" : "gaussian",
    "boron_peak_type" : "kubo_sakai",
    "background_type" : "linear",
    "overlap_rois" : True,
    "roi_width" : 15,
    "B_roi_width" : 20
}

#String-Object Map: used for UI interactions, makes a lot of things easier & allows for peaks/background types to be interchangeable
som = {
    "backgrounds":
    {
        "linear":LinearBackground,
        "quadratic":QuadraticBackground,
        "arctan" : ArctanBackground
    },
    "peaks":
    {
        "gaussian":GaussianPeak,
        "kubo_sakai":KuboSakaiBoronPeak
    }
}


from abc import ABC, abstractmethod

class Model(ABC):
    """Defines some standard functions for peaks/background models to rpevent code reuse"""
    def __init__(self):
        super().__init__()
    
    @staticmethod
    @abstractmethod
    def get_entry_fields(self):
        """Returns names of fields to be used when manually adding this peak/background model. 
        
        Example: ["Center (keV)", "Amplitude (cps)"] for Gaussian-type peaks.
        """
        pass

    @abstractmethod
    def get_type(self):
        """Returns the key to this peak/background class in the String-Object Map (constants.som)"""
        pass

    @abstractmethod
    def get_num_params(self):
        """Returns the number of parameters this model takes"""
        pass
    
    @abstractmethod
    def get_ydata(self, xdata):
        """Using the current model parameters, model the y values at every point in xdata, and return them"""
        pass
    @abstractmethod
    def get_ydata_with_params(self,xdata,params):
        """Using the provided parameters, model the y values at every point in xdata, and return them (used by the fitter)"""
        pass
    
    @abstractmethod
    def handle_entry(self, entry):
        """Handle user data entry. 
        
        The entry list is in the same order as the one provided by get_entry_fields()
        """
        pass
    @abstractmethod
    def __str__(self):
        pass
    
    #Getters and setters

    @abstractmethod
    def set_params(self, newParams):
        pass
    
    @abstractmethod
    def get_params(self):
        pass

    @abstractmethod
    def set_variances(self, variances):
        pass
    
    @abstractmethod
    def get_variances(self):
        pass

    @abstractmethod 
    def set_original_params(self, params):
        pass

    @abstractmethod
    def get_original_params(self):
        pass

    @abstractmethod
    def set_original_variances(self, vars):
        pass

    @abstractmethod
    def get_original_variances(self):
        pass

class Peak(Model):
    def __init__(self):
        super().__init__()
    
    @abstractmethod
    def get_area(self):
        """Returns the area under this peak"""
        pass
    
    @abstractmethod
    def get_area_stdev(self):
        """Returns the standard deviation of the area under this peak"""
        pass

    @abstractmethod
    def get_fwhm(self):
        """Returns a formatted FWHM width for the peak, or "N/A" if the peak doesn't have a FWHM"""

    #Getters and Setters

    @abstractmethod
    def get_ctr(self):
        pass
    

class StandardPeak(Peak):
    def __init__(self):
        super().__init__()

    @staticmethod
    @abstractmethod
    def guess_params(xdata, ydata):
        """Returns a list of StandardPeak objects representing all found peaks within a set of x-y data represented by xdata and ydata"""
        pass

class BoronPeak(Peak):
    def __init__(self):
        super().__init__()
        
    @staticmethod
    @abstractmethod
    def guess_params(xdata, ydata):
        """Returns a single BoronPeak object representing the closest match for a boron peak within the provided x-y data"""
        pass

    @staticmethod
    @abstractmethod
    def remove_from_data(xdata, ydata):
        """Returns cleaned ydata with the influence fo the boron peak subtracted from it.
        
        This is used because the Boron peak can often overshadow other peaks and make it harder for the peak finding function to work.
        Therefore, we subtract the boron peak from the data and find peaks in the resultant data.
        """
        pass
    

class Background(Model):
    def __init__(self):
        super().__init__()
    
    @staticmethod
    @abstractmethod
    def guess_params(xdata,ydata):
        """Returns a single Background object representing the closest match for a background within the provided x-y data"""
        pass


class Evaluator(ABC):
    """A class which takes ROIs containing peak matches and converts those peak matches to readable results."""
    def __init__(self):
        super().__init__()
    
    @staticmethod
    @abstractmethod
    def get_name():
        "Returns a name for the Evaluator"
        pass

    @staticmethod
    @abstractmethod
    def get_headings(self):
        """Return a list of column headings to be used with the results"""
        pass
    
    @abstractmethod
    def get_results(self):
        """Return a list of results from each ROI"""
        pass


from backend import ActivationAnalysis, ROI
from baseClasses import Model, Peak, Background, StandardPeak, BoronPeak
from evaluators import MassSensEval
from models import LinearBackground, QuadraticBackground, ArctanBackground, GaussianPeak, KuboSakaiBoronPeak
from parsers import SpectrumParser, StandardsFileParser, ExcelWriter, CSVWriter
from constants import som



import re
import os

import xylib
import numpy as np
from openpyxl import Workbook

from util import KnownPeak


class SpectrumParser:
    """Parser for all Spectrum file formats"""
    def __init__(self, fname):
        self.fname = fname
        if self.fname.split(".")[-1].lower() == "spe":
            self.speFile = True
        else:
            self.speFile = False
    def getValues(self):
        if self.speFile:
            #Reader specifically for SPE files from Genie
            with open(self.fname) as f:
                text = f.read()
            sections = dict(section.split('\n', 1) for section in text.split('$')[1:])
            livetime, realtime = sections["MEAS_TIM:"].strip().split()
            livetime = float(livetime)
            realtime = float(realtime)
            startind, endind = sections["DATA:"].split("\n")[0].split()
            data = sections["DATA:"].split("\n")[1:-1]
            data = [int(i) for i in data]
            intercept, slope = sections["ENER_FIT:"].strip().split()[:2]
            intercept = float(intercept)
            slope = float(slope)
            energies = [intercept + i*slope for i in range(int(startind),int(endind)+1)]
            cps = [total/livetime for total in data]
            return {"livetime":livetime, "realtime":realtime, "energies":np.array(energies), "cps":np.array(cps)}
        else:
            #Otherwise just use xylib
            d = xylib.load_file(self.fname, '')
            block = d.get_block(0)
            meta = block.meta
            metaDict = {meta.get_key(i) : meta.get(meta.get_key(i)) for i in range(meta.size())}
            ncol = block.get_column_count()
            nrow = block.get_point_count()
            data = [[block.get_column(i).get_value(j) for j in range(nrow)] for i in range(1, ncol+1)]
            livetime = metaDict["live time (s)"]
            realtime = metaDict["real time (s)"]
            return {"livetime":livetime, "realtime":realtime, "energies":np.array(data[0]), "cps":np.array(data[1])/float(livetime)}

class StandardsFileParser:
    """Parser for standards files with peak locations, sensitivities, etc."""
    def __init__(self,fname: str):
        self.fname = fname
        self.peaks = None
    def extract_peaks(self, delayed):
        if self.peaks != None:
            return self.peaks
        with open(self.fname) as f:
            lines = f.readlines()
        headings = re.sub(r'[^\x00-\x7F]+','', lines[0]).strip().split(",") #Eliminate non-ASCII characters
        lines = [line.split(",") for line in lines[1:]]
        try:
            peakIndex = headings.index("Energy (keV)")
            isoIndex = headings.index("Isotope")
        except:
            raise ValueError("Bad Sensitivity File Format.") # Those 2 headings are necessary
        #set of Regexes to test the other headings with
        reMass = re.compile(r"Mass \((\w+)\)")
        reSens = re.compile (r"Sensitivity \(cps/(\w+)\)")
        reHalfLife = re.compile(r"[Hh]alf-[Ll]ife \((\w+)\)")
        reDecayConstant = re.compile(r"[Dd]ecay [Cc]onstant \(1/(\w+)\)")

        #indexes to headings for each thing
        halfLifeIndex=None
        decayConstantIndex = None
        sensIndex = None
        massIndex = None

        #Booleans telling us whether half-life/decay constant is used and whether sensitivity/mass is used.
        decayConstant = None
        useSensitivity = None
        
        #Units for mass and decay time
        unit = None
        decayUnit = None
        for i,h in enumerate(headings):
            if reMass.match(h) != None:
                useSensitivity = False
                massIndex = i
                unit = reMass.match(h).group(1)
            elif reSens.match(h) != None:
                useSensitivity = True
                sensIndex = i
                unit = reSens.match(h).group(1)
            elif reDecayConstant.match(h):
                decayConstant = True
                decayUnit = reDecayConstant.match(h).group(1)
                decayConstantIndex = i
            elif reHalfLife.match(h):
                decayConstant = False
                decayUnit = reHalfLife.match(h).group(1)
                halfLifeIndex = i
        
        #Create the known peak lists
        if useSensitivity == None:
            self.peaks = [KnownPeak(l[isoIndex],float(l[peakIndex])) for l in lines]
        elif useSensitivity:
            self.peaks = [KnownPeak(l[isoIndex],float(l[peakIndex]), sensitivity = float(l[sensIndex]), unit=unit) for l in lines]
        else:
            self.peaks = [KnownPeak(l[isoIndex],float(l[peakIndex]), mass = float(l[massIndex]), unit=unit) for l in lines]
        
        #Adding the NAA parameters if needed
        if delayed:
            if decayConstant == None:
                raise ValueError("Delayed analysis, must provide half-life or decay constant in sensitivity file")
            elif decayConstant:
                for i in range(len(self.peaks)):
                    self.peaks[i].set_NAA_params(decayConstant = lines[i][decayConstantIndex], unit=decayUnit)
            else:
                for i in range(len(self.peaks)):
                    self.peaks[i].set_NAA_params(halfLife = lines[i][halfLifeIndex], unit=decayUnit)
        return self.peaks

class CSVWriter:
    """Writes a CSV file with name fname to the results/projectID directory.
    
    The first row of the file contains column headings specified in headings.
    All remaining data is specified in data.
    """
    def __init__(self, projectID, fname, headings, data):
        self.fname = "./results/" + projectID + "/" + fname
        self.headings = headings
        self.data = data
    def write(self):
        with open(self.fname, "w") as f:
            f.seek(0)
            f.write(",".join(self.headings)+"\n")
            for line in self.data:
                try:
                    ld = [str(e) for e in line[0]]
                except:
                    ld = [str(e) for e in line]
                f.write(','.join(ld)+"\n")

class ExcelWriter:
    def __init__(self, projectID, projectTitle, allFilenames, headings, data):
        """Writes results from a project into an Excel File."""
        self.fname = "./results/" + projectID + "/" + projectTitle.replace(" ","_").replace("\\","").replace("/","") + ".xlsx"
        self.allFilenames = allFilenames
        self.headings = headings
        self.data = data
    def write(self):
        wb = Workbook()
        
        #First sheet w/compiled data
        ws = wb.active
        ws.title = "All Files"
        ws["A1"] = "Filename"
        for i in range(len(self.headings[0][0])):
            _ = ws.cell(row=1, column=i+2, value=self.headings[0][0][i])
        rowCount = 1
        for i in range(len(self.allFilenames)):
            for j in range(len(self.data[i][0])): #its data[i][0] and not data[i] because this writer only cares about 1 evaluator, the Mass/Sens one used in the program.
                _ = ws.cell(row=rowCount+1, column=1, value=os.path.split(self.allFilenames[i])[1])
                for k in range(len(self.data[i][0][j])):
                    _ = ws.cell(row=rowCount+1, column=k+2, value=self.data[i][0][j][k])
                rowCount += 1

        #remaining sheets
        for i in range(len(self.allFilenames)):
            newWs = wb.create_sheet(os.path.split(self.allFilenames[i])[1][:31]) #sheet name can't be longer than 31 chars
            for l in range(len(self.headings[0][0])):
                _ = newWs.cell(row=1, column=l+1, value=self.headings[0][0][l])
            for j in range(len(self.data[i][0])): #its data[i][0] and not data[i] because this writer only cares about 1 evaluator, the Mass/Sens one used in the program.
                for k in range(len(self.data[i][0][j])):
                    _ = newWs.cell(row=j+2, column=k+1, value=self.data[i][0][j][k])
        
        wb.save(self.fname)






from setuptools import setup
setup(
    name="openags",
    version='0.1.1',
    packages=['openags'],
    install_requires=['numpy','scipy','sigfig','xylib-py-wheels'],
    classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          'Topic :: Scientific/Engineering :: Physics',
          'Topic :: Software Development :: Libraries :: Python Modules',
          ],
      author='Christopher Stallard',
      author_email='christopher.stallard1@gmail.com',
      license='MIT',
      url='https://github.com/chris-stallard1/OpenAGS'
      )


import os
import warnings
import itertools

import numpy as np
from scipy.optimize import curve_fit, OptimizeWarning
from sigfig import round

from util import multiple_peak_and_background, get_curve, binary_search_find_nearest, set_all_params, KnownPeak
from constants import default_prefs, som
from parsers import SpectrumParser, StandardsFileParser, CSVWriter, ExcelWriter
from models import GaussianPeak, LinearBackground



class ActivationAnalysis:
    """An application class representing the whole backend. 
    
    This is the ONLY backend class that the frontend should interact with.

    ----Variables----
    #User Preferences
    userPrefs: Dict[str, any] = None
    #Extracted contents and metadata from project files
    fileData: List[Dict[str, any]] = None
    #List of spectrum filenames, data from each correlates to fileData
    fileList: List[str] = None
    #dictionary mapping peak location to KnownPeak onjects
    knownPeaks: Dict[float, KnownPeak] = None
    #Regions of Interest, see ROI class below
    ROIs: List[ROI] = None
    #Isotopes being analyzed
    isotopes: List[str] = None
    #Project Title
    title: str
    #Whether all regions of interest for this project have been fitted
    ROIsFitted: bool
    #Whether evaluators have been run and results generated for this project
    resultsGenerated: bool
    #Whether or not this is a Delayed Gamma Analysis
    delayed: bool"""
    
    def __init__(self, userPrefs = default_prefs, title = ""):
        self.userPrefs = userPrefs
        self.fileData = None
        self.fileList = None
        self.knownPeaks = None
        self.ROIs = None
        self.isotopes = None
        self.title = title
        self.ROIsFitted = False
        self.resultsGenerated = False
        self.delayed = False

    def load_from_dict(self, stored_data):
        """Sets variables for an analysis object based on a dictionary exported by the export_to_dict() function."""

        if "userPrefs" in stored_data: #otherwise keep default prefs
            self.userPrefs = stored_data["userPrefs"]
        
        self.title = stored_data["title"]
        self._load_spectra(stored_data["files"])
        self._load_known_peaks(stored_data["standardsFilename"])
        energies = self.fileData[0]["energies"]
        cps = self.fileData[0]["cps"]

        self.ROIsFitted = stored_data["ROIsFitted"]
        self.ROIs = []
        for ROIData in stored_data["ROIs"]:
            lowerIndex = ROIData["indicies"][0]
            upperIndex = ROIData["indicies"][1]
            r = ROI(energies[lowerIndex:upperIndex], cps[lowerIndex:upperIndex], [lowerIndex, upperIndex])
            r.load_from_dict(ROIData)
            self.ROIs.append(r)

        self.isotopes = list(set(itertools.chain(*[r.get_isotopes() for r in self.ROIs])))
        self.resultsGenerated = stored_data["resultsGenerated"]
        if self.resultsGenerated:
            for i in range(len(stored_data["results"])):
                self.fileData[i]["results"] = stored_data["results"][i]
                self.fileData[i]["resultHeadings"] = stored_data["resultHeadings"]
                self.fileData[i]["evaluatorNames"] = stored_data["evaluatorNames"]

        self.delayed = stored_data["delayed"]
        if self.delayed and "NAATimes" in stored_data:
            for i in range(len(stored_data["NAATimes"])):
                self.fileData[i]["NAATimes"] = stored_data["NAATimes"][i]


    def export_to_dict(self):
        """Exports the current project state to a dictionary"""
        exportROIs = [r.export_to_dict() for r in self.ROIs]
        outDict =  {
            "userPrefs" : self.userPrefs,
            "title" : self.title,
            "files" : self.fileList,
            "standardsFilename" : self.standardsFilename,
            "ROIsFitted" : self.ROIsFitted,
            "ROIs" : exportROIs,
            "resultsGenerated" : self.resultsGenerated,
            "delayed" : self.delayed
        }

        if self.resultsGenerated:
            outDict["results"] = [fd["results"] for fd in self.fileData]
            outDict["resultHeadings"] = self.fileData[0]["resultHeadings"]
            outDict["evaluatorNames"] = self.fileData[0]["evaluatorNames"]

        if self.delayed and "NAATimes" in self.fileData[0]:
            outDict["NAATimes"] = [fd["NAATimes"] for fd in self.fileData]
            
        return outDict

    def _load_spectra(self, files):
        """Parse and add the spectrum files specified in the files list"""
        self.fileList = files
        self.fileData = [SpectrumParser(f).getValues() for f in files]

    def _load_known_peaks(self, standardsFilename):
        """Parse and add known peaks from a standards file"""
        self.standardsFilename = standardsFilename
        self.knownPeaks = {}
        peakList = StandardsFileParser(standardsFilename).extract_peaks(self.delayed)
        for p in peakList:
            c = p.get_ctr() # avoid collisions by changing the center by .01 eV in this dictionary, without affecting the actual object
            while c in self.knownPeaks:
                c += 0.00001
                
            self.knownPeaks[c] = p

    def update_ROIs(self, addedIsotopes, removedIsotopes = []):
        """Update our ROIs, adding some isotopes and potentially removing some.
        
        This function can be used to create ROIs by calling it with only 1 argument.
        """

        if not addedIsotopes and not removedIsotopes:
            #If we are told to do nothing, do nothing
            return

        #If we are doing anything, the new ROIs we create might not be fitted.
        self.ROIsFitted = False

        #ensure no duplicate additions
        addedIsotopes = [iso for iso in addedIsotopes if iso not in self.isotopes and iso in self.get_all_isotopes()]
        self.isotopes.extend(addedIsotopes)

        #remove isotopes that the user wants to remove
        for iso in removedIsotopes:
            try:
                self.isotopes.remove(iso)
            except ValueError: #if there is a duplicate, move on
                pass
        
        #either remove ROIs completely or add to an "edit list" if some, but not all, isotopes in roi have been removed
        editList, ROIs = [], []
        for r in self.ROIs:
            isotopes = r.get_isotopes()
            filtered = [iso for iso in isotopes if iso not in removedIsotopes]
            if filtered:
                # some isotopes remaining after remove, so keep the ROI.
                ROIs.append(r)
                if len(filtered) < len(isotopes):
                    editList.extend(kp.get_ctr() for kp in r.get_known_peaks)
        self.ROIs = ROIs

        regions = []
        peaks = []
        otherPeaks = []

        #create new ROIs
        sortedKeys = sorted(self.knownPeaks.keys())

        for k in sortedKeys:
            print(k)
            p = self.knownPeaks[k]
            if p.get_ele() in addedIsotopes or p.get_ctr() in editList:
                if p.get_ele() == "B-11" and p.get_ctr() < 480 and p.get_ctr() > 470: #special case for boron
                    lowerBound = max(p.get_ctr() - self.userPrefs["B_roi_width"], 0)
                    upperBound = min(p.get_ctr() + self.userPrefs["B_roi_width"], self.fileData[0]["energies"][-1])
                else:
                    lowerBound = max(p.get_ctr() - self.userPrefs["roi_width"], 0)
                    upperBound = min(p.get_ctr() + self.userPrefs["roi_width"], self.fileData[0]["energies"][-1])
                
                regions.append(lowerBound)
                regions.append(upperBound)
                
                peaks.append([p])
                lowerIndex = binary_search_find_nearest(sortedKeys, lowerBound)
                upperIndex = binary_search_find_nearest(sortedKeys, upperBound)
                otherPeaks.append([self.knownPeaks[e] for e in sortedKeys[lowerIndex:upperIndex]])

        if self.userPrefs["overlap_rois"]:
            i=0
            while i < len(regions) - 1:
                if regions[i] > regions[i+1]: #if there is an overlap, delete both points that overlap, leaving a single, larger region
                    del regions[i]
                    del regions[i]
                    peaks[i//2] += peaks[i//2+1]
                    del peaks[i//2+1]
                    otherPeaks[i//2] += otherPeaks[i//2+1]
                    del otherPeaks[i//2+1]
                else:
                    i += 1

        energies = self.fileData[0]["energies"]
        cps = self.fileData[0]["cps"]
        print(regions)
        for i in range(0,len(regions),2):
            lowerIndex = binary_search_find_nearest(energies, regions[i])
            upperIndex = binary_search_find_nearest(energies, regions[i+1])
            boronRegion = "B-11" in [p.get_ele() for p in peaks[i//2]] and not self.delayed and regions[i] < 477.6 and regions[i+1] > 477.6
            r = ROI(energies[lowerIndex:upperIndex],cps[lowerIndex:upperIndex], [lowerIndex, upperIndex], boronRegion, self.userPrefs)
            r.set_known_peaks(peaks[i//2], otherPeaks[i//2])
            self.ROIs.append(r)
        self.ROIs = sorted(self.ROIs, key=lambda x:x.get_range()[0])

    def fit_ROIs(self):
        """Fits all ROIs that aren't fitted: convenience function that calls several functions on each unfitted ROI."""
        for ROI in self.ROIs:
            if not ROI.fitted:
                ROI.add_peaks()
                ROI.add_bg()
                ROI.fit()
        self.ROIsFitted = True
        return self.ROIs

    def get_entry_repr(self, model, name, ROIIndex, params):
        """Get Entry Represenaation for provided entry, given peak and background type."""
        if model == "peaks":
            testObj = som[model][name]()
            testObj.handle_entry(params, bounds=self.ROIs[ROIIndex].get_range())
            return str(testObj), testObj.get_params()
        elif model == "backgrounds":
            tmpObj = som[model][name].guess_params(self.ROIs[ROIIndex].get_energies(), self.ROIs[ROIIndex].get_cps())
            return str(tmpObj), tmpObj.get_params()

    def set_ROI_range(self, ROIIndex, newRange):
        """Set the range (of energy values) of the ROI at index ROIIndex to the values in values"""
        energies = self.fileData[0]["energies"]
        cps = self.fileData[0]["cps"]
        lowerIndex = binary_search_find_nearest(energies, newRange[0])
        upperIndex = binary_search_find_nearest(energies, newRange[1])
        self.ROIs[ROIIndex].set_data([energies[lowerIndex], energies[upperIndex]], energies[lowerIndex:upperIndex], cps[lowerIndex:upperIndex], [lowerIndex, upperIndex])

    def run_evaluators(self, evaluators, e_args):
        """Run a list of evaluators on our ROIs, with arguments specified in the list e_args"""
        ROIsToEval = [r for r in self.ROIs if r.fitted]
        for i in range(len(self.fileData)):
            if i != 0:
                energies = self.fileData[i]["energies"]
                cps = self.fileData[i]["cps"]
                for r in ROIsToEval:
                    if self.delayed:
                        for kp in r.get_known_peaks():
                            kp.set_delay_times(*self.fileData[i]["NAATimes"], self.fileData[i]["realtime"]/60)
                    bounds = r.get_range()
                    lowerIndex = binary_search_find_nearest(energies, bounds[0])
                    upperIndex = binary_search_find_nearest(energies, bounds[1])
                    r.reanalyze(energies[lowerIndex:upperIndex], cps[lowerIndex:upperIndex])
            self.fileData[i]["results"] = [e(ROIsToEval).get_results(*args) for e, args in zip(evaluators, e_args)]
            self.fileData[i]["resultHeadings"] = [e.get_headings(ROIsToEval[0]) for e in evaluators]
            self.fileData[i]["evaluatorNames"] = [e.get_name() for e in evaluators]
        self.resultsGenerated = True

    def write_results_file(self, projectID, filename):
        """Writes a results file, format/spec depends on the filename of the request.

        Implements ExcelWriter to write results and CSVWriter to write results or spectrum file data.
        """
        if filename.split(".")[-1] == "xlsx":
            headings = [fd["resultHeadings"] for fd in self.fileData]
            data = [fd["results"] for fd in self.fileData]
            ew = ExcelWriter(projectID, self.get_title(), self.fileList, headings, data)
            ew.write()
        elif filename[-21:] == "_Analysis_Results.csv":
            origFilename = filename.replace("_Analysis_Results.csv","")
            for filename, fileData in zip(self.fileList, self.fileData):
                if os.path.split(filename)[1].split('.')[0] == origFilename:
                    cw = CSVWriter(projectID, filename, fileData["resultHeadings"][0], fileData["results"])
                    cw.write()
                    break
        elif filename[-7:] == "_xy.csv":
            origFilename = filename.replace("_xy.csv","")
            for filename, fileData in zip(self.fileList, self.fileData):
                if os.path.split(filename)[1].split('.')[0] == origFilename:
                    cw = CSVWriter(projectID, filename, ["Energy (keV)", "Counts Per Second"], zip(fileData["energies"], fileData["cps"]))
                    cw.write()
                    break
    #Getters and Setters
    def set_delayed_times(self, i, irr, wait, count):
        self.fileData[i]["NAATimes"] = [irr, wait, count]
        
    def get_all_isotopes(self):
        return set(v.get_ele() for k, v in self.knownPeaks.items())
    
    def get_known_annots(self):
        return [[[kp.get_ctr(), kp.get_ele()] for kp in r.peaksInRegion] for r in self.ROIs]
    
    def get_naa_times(self):
        return [fd["NAATimes"] for fd in self.fileData]
    
    def get_unfitted_ROIs(self):
        return [i for i, ROI in enumerate(self.ROIs) if not ROI.fitted]
    
    def set_user_prefs(self, newPrefs):
        self.userPrefs.update(newPrefs)

    def get_known_peaks(self):
        return self.knownPeaks
    
    def get_title(self):
        return self.title
    
    def set_title(self, newTitle):
        self.title = newTitle

    def get_isotopes(self):
        return self.isotopes
        
    def get_filename_list(self):
        return [os.path.split(f)[1] for f in self.fileList]

    def get_all_entry_fields(self):
        return {
            "peaks" : {k : v.get_entry_fields() for k, v in som["peaks"].items()},
            "backgrounds" : {k : v.get_entry_fields() for k, v in som["backgrounds"].items()}
        }


class ROI:
    def __init__(self, energies, cps, indicies, boronROI = False, userPrefs = default_prefs):
        self.energies = energies
        self.range = (energies[0], energies[-1])
        self.cps = cps
        self.knownPeaks = []
        self.peaksInRegion = []
        self.userPrefs = userPrefs
        self.indicies = indicies
        self.peaks = []
        self.bg = None
        self.peakPairs = None
        self.fitted = False
        self.boronROI = boronROI

    def load_from_dict(self, stored_data):
        """Sets variables for an ROI object based on a dictionary exported by the export_to_dict() function."""
        if "peaks" in stored_data:
            self.peaks = [som["peaks"][p["type"]](*p["params"], variances=p["variances"]) for p in stored_data["peaks"]]
            self.bg = som["backgrounds"][stored_data["background"]["type"]](*stored_data["background"]["params"],variances=stored_data["background"]["variances"])
            self.fitted = (self.peaks[0].get_variances()[0] != None)
        else:
            self.fitted = False
        
        self.knownPeaks = [KnownPeak().load_from_dict(kp) for kp in stored_data["knownPeaks"]] 
        if "peakPairs" in stored_data:
            self.peakPairs = self.originalPeakPairs = [(self.peaks[i], self.knownPeaks[j]) for i, j in stored_data["peakPairs"]]
    def export_to_dict(self):
        """Exports the current ROI state to a dictionary"""
        PIR = [p.export_to_dict() for p in self.peaksInRegion]
        exportKnownPeaks = [kp.export_to_dict() for kp in self.knownPeaks]
        outDict = {
                "indicies" : self.indicies,
                "knownPeaks" : exportKnownPeaks,
                "peaksInRegion" : PIR
        }

        if self.bg != None:
            outDict["peaks"] = [{"type" : p.get_type(), "params" : p.get_original_params(), "variances": p.get_original_variances()} for p in self.peaks]
            outDict["background"] = {"type" : self.bg.get_type(), "params" : self.bg.get_original_params(), "variances": self.bg.get_original_variances()}
        if self.peakPairs != None:
            outDict["peakPairs"] = [(self.peaks.index(p), self.knownPeaks.index(kp)) for p, kp in self.originalPeakPairs]
        return outDict

    def add_peaks(self):
        """Find and add peaks to own model (guesss params)"""
        if self.boronROI:
            BPeak = som["peaks"][self.userPrefs["boron_peak_type"]]
            self.peaks = [BPeak.guess_params(self.energies, self.cps)]
            scrubbedCPS = BPeak.remove_from_data(self.energies, self.cps)
            print(scrubbedCPS)
            self.peaks += som["peaks"][self.userPrefs["peak_type"]].guess_params(self.energies, scrubbedCPS)
        else:
            self.peaks = som["peaks"][self.userPrefs["peak_type"]].guess_params(self.energies, self.cps)
   
    def add_bg(self):
        """Find and add background to own model (guesss params)"""
        self.bg = som["backgrounds"][self.userPrefs["background_type"]].guess_params(self.energies, self.cps)

    def fit(self, reanalyze = False):
        """Fit our model to the data within the ROI, using the guessed params as initial ones"""
        f = lambda x,*params: multiple_peak_and_background(self.peaks, self.bg, x, params)
        p0 = np.array(self.bg.get_params() + list(itertools.chain.from_iterable([p.get_params() for p in self.peaks])))
        with warnings.catch_warnings():
            warnings.simplefilter("error", OptimizeWarning)
            try:
                params, cov = curve_fit(f, self.energies, self.cps, p0=p0)
                variances = np.diag(cov)
                set_all_params(self.peaks, self.bg, params, variances, reanalyze)
                self.fitted = True
            except Exception as e:
                print(e)
                self.fitted = False
                pass
    
    def get_fitted_curve(self, xdata = None):
        """Get the output of our fit (ydata) given x values"""
        if xdata == None:
            xdata = np.arange(self.range[0], self.range[-1], .01)
        return [list(xdata), get_curve(self.peaks, self.bg, xdata), list(self.bg.get_ydata(xdata))]

    def get_closest_peak(self, peak):
        if len(self.peaks) == 0:
            return None
        target = peak.get_ctr()
        distance = [abs(p.get_ctr() - target) for p in self.peaks]
        index = np.argmin(distance)
        return self.peaks[index]
    
    def set_original_peak_pairs(self, energyPairs):
        """Original peak pairs are set so that they don't change when the ROI is reanalyzed on new data and can be exported/imported easily."""
        pairs = []
        peakCtr = np.array([p.get_ctr() for p in self.peaks])
        knownCtr = np.array([p.get_ctr() for p in self.knownPeaks])
        nearest_peak = lambda target: self.peaks[np.argmin(abs(peakCtr-target))]
        nearest_known = lambda target: self.knownPeaks[np.argmin(abs(knownCtr-target))]
        pairs = [(nearest_peak(p), nearest_known(kp)) for p, kp in energyPairs]
        self.peakPairs = self.originalPeakPairs = pairs

    def reanalyze(self, newEnergies, newCPS):
        """Re-runs the fit on a new set of energies and cps from another spectrum file, and re-match peaks"""
        if self.peakPairs == None:
            raise RuntimeError("Reanalyze called before peak pairs created!")
        self.energies = newEnergies
        self.cps = newCPS
        self.fit(True)
        outputs = []
        peakCtrs = np.array([p.get_ctr() for p in self.peaks])
        for peak, knownPeak in self.originalPeakPairs:
            target = peak.get_ctr()
            closestMatch = self.peaks[np.argmin(abs(peakCtrs - target))]
            outputs.append([closestMatch, knownPeak])  
        self.peakPairs = outputs      
        return outputs
    
    #Getters and Setters
    def set_peaks(self, peaks):
        self.peaks = peaks
    
    def get_peaks(self):
        return self.peaks

    def get_isotopes(self):
        return [kp.get_ele() for kp in self.knownPeaks]

    def get_peak_ctrs(self):
        return [p.get_ctr() for p in self.peaks]

    def get_known_peaks(self):
        return self.knownPeaks
    
    def get_range(self):
        return list(self.range)

    def get_formatted_range(self):
        return [str(round(float(self.range[0]), decimals=1)), str(round(float(self.range[1]), decimals=1))]

    def set_range(self, newRange):
        self.range = newRange
        self.energies = np.arange(newRange[0], newRange[1], .01)

    def get_energies(self):
        return list(self.energies)

    def get_cps(self):
        return list(self.cps)

    def set_data(self, newRange, energies, cps, indicies):
        self.range = newRange
        self.energies = energies
        self.cps = cps
        self.indicies = indicies

    def set_known_peaks(self, peaks, otherPeaks):
        self.knownPeaks = peaks
        self.peaksInRegion = otherPeaks

    def set_background(self, bg):
        self.bg = bg

    def get_background(self):
        return self.bg

    def get_peak_pairs(self):
        return self.peakPairs

    def get_indicies(self):
        return self.indicies








from abc import ABC, abstractmethod
import math
import itertools

import numpy as np
from numpy import NaN, inf, arange, asarray
import xylib
from sigfig import round

class KnownPeak:
    def __init__(self, elementName="None", center=0.0, unit = "mg", sensitivity=None, mass=None, halfLife = None, decayConstant = None, decayUnit="min"):
        """Represents a peak generated by isotope "elementName" with a predetermined energy centroid and possbly other parameters.
        
        The peak can have a known sensitivity or mass attached to it. This will be the divisor which the area is divided by at the end of the analysis.
        If and only if the analysis is NAA, the peak will also have a half-life or decay constant, representing that of the isotope that is producing it.
        These are used for time correction in the get_tcf() function.
        """
        self.elementName = elementName
        self.center = center
        self.delayed = True


        if sensitivity != None and mass != None:
            raise TypeError("Please provide no more than 1 of the following: mass, sensitivity")
        if mass != None:
            self.divisor = mass
            self.output = "Sensitivity ("+unit+")"
        elif sensitivity != None:
            self.divisor = sensitivity
            self.output = "Mass ("+unit+")"
        else:
            self.divisor = None
            self.output = "Peak Area (cps)"
        
        if halfLife != None and decayConstant != None:
            raise TypeError("Please provide no more than 1 of the following: half-life, decay constant")
        elif halfLife != None:
            self.decayConstant = math.log(2) / halfLife
        elif decayConstant != None:
            self.decayConstant = decayConstant
        else:
            self.delayed = False

        if self.delayed:
            #these seem reversed because decay constant unit is 1/time
            if decayUnit in ("s", "sec"):
                self.decayConstant *= 60
            if decayUnit in ("h", "hr"):
                self.decayConstant /= 60
    
    #Input Functions: load data into object

    def load_from_dict(self, d):
        """Takes in a dictionary represenation of this object (generated by export_to_dict()), and sets parameters accordingly."""
        self.elementName = d["ele"]
        self.center = d["ctr"]

        if "divisor" in d:
            self.divisor = d["divisor"]
            self.output = d["output"]
        
        if "decayConstant" in d:
            self.decayConstant = d["decayConstant"]
        return self

    #Interesting functions: Do more than get or set but aren't inputs/outputs

    def get_tcf(self):
        """Returns the time correction factor to divide by: 1 for PGAA, and dependent on irradiation, wait, and count time for NAA"""
        if self.delayed:
            return (1 - math.e ** (-1 * self.decayConstant * self.irrTime)) * math.e ** (-1 * self.decayConstant * self.waitTime) * (1 - math.e ** (-1 * self.decayConstant * self.countTime))
        else:
            return 1
    

    #Getters and setters

    def set_delay_times(self, irr, wait, count):
        self.irrTime = irr
        self.waitTime = wait
        self.countTime = count

    def get_ctr(self):
        return self.center

    def get_ele(self):
        return self.elementName

    def get_output(self):
        return self.output
    
    def set_NAA_params(self, halfLife=None, decayConstant=None, unit="min"):
        """Set certain parameters for NAA. 
        
        This mirrors some functionality in __init__ but helps to avoid a certain 
        tangled mess of if cases that used to be in parsers.py
        """
        if halfLife != None and decayConstant != None:
            raise TypeError("Please provide no more than 1 of the following: half-life, decay constant")
        elif halfLife != None:
            self.decayConstant = math.log(2) / halfLife
        elif decayConstant != None:
            self.decayConstant = decayConstant
        else:
            raise TypeError("Either half-life or decay constant should be provided")

        if unit in ("s", "sec"):
            self.decayConstant *= 60
        elif unit in ("h", "hr"):
            self.decayConstant /= 60

    #Output Methods: Transform the data in the object into something

    def export_to_dict(self):
        """Exports all information about this peak to a dictionary, from which the peak can later be loaded."""
        outDict = {
                "ele" : self.elementName,
                "ctr" : self.center
                }
        
        if self.divisor != None:
            outDict["divisor"] = self.divisor
            outDict["output"] = self.output

        if self.delayed:
            outDict["decayConstant"] = self.decayConstant
        
        return outDict

    def __str__(self):
        """Returns a string representation of this peak as its element name concatenated with its pek centroid."""
        return self.elementName + " : " + str(round(float(self.center), decimals=1))
    
    def get_results(self, area, areaStdev):
        """Combines this peak (with known sensitivity or mass value) with a peak fround in the data to get the mass/sensitivity of that peak.

        The peak in the data is fitted and found to have area "area" and area std. dev. "areaStdev". It is then matched to this known peak,
        and this method returns the desired output for the original peak given that match.
        """
        if self.divisor == None:
            return [area/self.get_tcf(), areaStdev/self.get_tcf()]
        return [area/self.divisor/self.get_tcf(), areaStdev/self.divisor/self.get_tcf()]

def multiple_peak_and_background(peaks, background, x, params):
    """Function used for curve_fit, splits the parameters between the component peaks and background.
    
    Uses get_num_params() function of peaks/backgrounds in order to determine how many paramaters to allocate them.
    Computes y values for a given set of parameters and returns them.
    """
    y = np.zeros_like(x)
    leftCounter = 0
    rightCounter = background.get_num_params()
    y += background.get_ydata_with_params(x,params[leftCounter:rightCounter])
    
    for peak in peaks:
        leftCounter = rightCounter
        rightCounter = leftCounter + peak.get_num_params()
        y += peak.get_ydata_with_params(x,params[leftCounter:rightCounter])
        
    return y

def set_all_params(peaks, background, params, variances, reanalyze):
    """Sets parameters of background/peaks to the result of the fitter

    If the data is being fit for the first time (not in a batch), original params and variances are set. 
    These are the parameters stored in the state file, so that the reloaded fits match the file the user adjusts the fits on.
    If this step wasn't taken, the program would apply the fit from the last file to the data from the first file.
    """
    leftCounter = 0
    rightCounter = background.get_num_params()

    noVars = (variances[0] == inf) or (variances[0] == -1 * inf) #sometimes, variances cannot be determined

    background.set_params(params[leftCounter:rightCounter])
    if not noVars:
        background.set_variances(variances[leftCounter:rightCounter])
    if not reanalyze:
        background.set_original_params(params[leftCounter:rightCounter])
        if not noVars:
            background.set_original_variances(variances[leftCounter:rightCounter])
    for peak in peaks:
        leftCounter = rightCounter
        rightCounter = leftCounter + peak.get_num_params()
        peak.set_params(params[leftCounter:rightCounter])
        if not noVars:
            peak.set_variances(variances[leftCounter:rightCounter])
        if not reanalyze:
            peak.set_original_params(params[leftCounter:rightCounter])
            if not noVars:
                peak.set_original_variances(variances[leftCounter:rightCounter])

def get_curve(peaks, background, x):
    """Sums peaks and background counts for a given set of X values"""
    y = np.zeros_like(x)
    y += background.get_ydata(x)
    for peak in peaks:
        y += peak.get_ydata(x)
    return list(y)

def binary_search_find_nearest(l, e):
    """Find the index of the element in list l which is closest to element e."""
    upperBound = len(l)
    lowerBound = 0
    guess = (upperBound + lowerBound)//2

    #edge cases, where len(l) <= 2
    if guess == 0:
        return guess
    if guess == 1:
        if abs(e - l[0]) < abs(e - l[1]):
            return 0
        return 1

    #Binary Search    
    while not (e < l[guess+1] and e > l[guess-1]):
        if e > l[guess]:
            lowerBound = guess + 1
        else:
            upperBound = guess - 1
        guess = (upperBound + lowerBound)//2
        if guess <= 2 or guess >= len(l)-2:
            break
    if e > l[guess]:
        guess += 1
    return guess

def ivw_combine(meas, stdev = None, variance = None):
    """Combines measurements through Inverse Variance Weighting.

    Arguments: meas -- The measurements to combine
    Keyword Arguments: stdev -- Standard deviations of the measurements
                       variance -- variances of the measurements
    (only stdev XOR variance should be provided)
    """
    var = None
    res = None
    if variance != None:
        var = 1/sum([1/v for v in variance])
        res = var * sum([m/v for m,v in zip(meas, variance)])
    else:
        var = 1/sum([1/s**2 for s in stdev])
        res = var * sum([m/s**2 for m,s in zip(meas, stdev)])
    return [res, var**.5]
